import itertools
from abc import ABCMeta
from abc import abstractmethod
from dataclasses import dataclass
from decimal import Decimal
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union


@dataclass
class Metric:
    """
        A metric is a contextualized numerical value with contextualized meaning that it has a name,
        a lineage (i.e., a set of ancestors) and tags on each ancestor.
    """
    name: str
    value: Union[int, float, Decimal]
    tags: Dict[str, str]  # The first value in the list is considered to be the parent


class MetricsTree(metaclass=ABCMeta):
    @abstractmethod
    def to_metrics(self, tag_context: Dict[str, str]) -> List[Metric]:
        pass


@dataclass
class MetricsNone(MetricsTree):
    def to_metrics(self, tag_context: Dict[str, str]) -> List[Metric]:
        return []


@dataclass
class MetricsLeaf(MetricsTree):
    name: str
    value: Union[int, float, Decimal]

    def to_metrics(self, tag_context: Dict[str, str]) -> List[Metric]:
        return [Metric(self.name, self.value, tag_context)]


@dataclass
class MetricsTags(MetricsTree):
    tags: Dict[str, str]
    internal: MetricsTree

    def to_metrics(self, tag_context: Dict[str, str]) -> List[Metric]:
        tag_context = tag_context.copy()
        tag_context.update(self.tags)
        return self.internal.to_metrics(tag_context)


@dataclass
class MetricsBranch(MetricsTree):
    children: List[MetricsTree]

    def to_metrics(self, tag_context: Dict[str, str]) -> List[Metric]:
        return [m for child in self.children for m in child.to_metrics(tag_context)]



"""\
pyjack
******

pyjack is debug/test/monkey-patching toolset that allows you to reversibly
replace all references to a function /object in memory with a
proxy function/object.

:copyright: Copyright 2009-2012 by Andrew Carter <andrewjcarter@gmail.com>
:license: MIT (http://www.opensource.org/licenses/mit-license.php)

"""
import sys as _sys
import gc as _gc
import types as _types
import inspect as _inspect

_WRAPPER_TYPES = (type(object.__init__), type(object().__init__),)


def proxy0(data):
    def proxy1(): return data

    return proxy1


if _sys.version_info[0] < 3:
    _CELLTYPE = type(proxy0(None).func_closure[0])
else:
    _CELLTYPE = type(proxy0(None).__closure__[0])


class PyjackException(Exception): pass


def connect(fn, proxyfn):
    """\
    :summary: Connects a filter/callback function to a function/method.

    :param fn: The function which to pyjack.
    :type fn:  :class:`types.FunctionType`, :class:`types.MethodType`, 
               :class:`types.BuiltinFunctionType`, :class:`BuiltinMethodType` 
               or any callable that implements :func:`__call__`

    :param proxyfn: Any callable.  It will be passed the original :attr:`fn`
                    and then any args, kwargs that were passed to the 
                    original :attr:`fn`. 
    :type proxyfn:  callable. 

    :returns: The new pyjacked function.  Note, this function object
              has a :func:`restore` that can be called to remove the
              pyjack filters/callbacks. 

    :raises: :class:`PyjackException`
    """
    fn_type = type(fn)
    if issubclass(fn_type, _types.FunctionType):
        return _PyjackFuncCode(fn, proxyfn)._fn
    elif issubclass(fn_type, _types.MethodType):
        return _PyjackFuncCode(fn.im_func, proxyfn)._fn
    elif issubclass(fn_type, (_types.BuiltinFunctionType,
                              _types.BuiltinMethodType,
                              type)):
        return _PyjackFuncBuiltin(fn, proxyfn)
    elif _sys.version_info < (2, 5) and issubclass(fn_type, type(file)):
        # in python 2.4, open is of type file, not :class:`types.FunctionType`
        return _PyjackFuncBuiltin(fn, proxyfn)
    elif issubclass(fn_type, _WRAPPER_TYPES):
        raise PyjackException("Wrappers not supported. Make a concrete fn.")
    elif isinstance(getattr(fn, '__call__', None), _types.MethodType):
        _PyjackFuncCode(fn.__call__.im_func, proxyfn)

        def restore():
            fn.__call__.im_func.restore()
            delattr(fn, 'restore')

        fn.restore = restore
        return fn
    else:
        bundle = (fn, fn_type,)
        raise PyjackException("fn %r of type '%r' not supported" % bundle)


def restore(fn):
    """\
    :summary:  Fully restores function back to original state. 

    :param fn: The pyjacked function returned by :func:`connect`. 

    .. note::

       Any pyjacked function has a :func:`restore` method, too.  So you can 
       call that instead of this procedural function -- it's up to you.     

    """
    fn.restore()


# SOMETHING BROKEN IN TYPES MODULE


def replace_all_refs(org_obj, new_obj):
    """
    :summary: Uses the :mod:`gc` module to replace all references to obj
              :attr:`org_obj` with :attr:`new_obj` (it tries it's best, 
              anyway). 

    :param org_obj: The obj you want to replace. 

    :param new_obj: The new_obj you want in place of the old obj. 

    :returns: The org_obj

    Use looks like:

    >>> import pyjack
    >>> x = ('org', 1, 2, 3)
    >>> y = x
    >>> z = ('new', -1, -2, -3)
    >>> org_x = pyjack.replace_all_refs(x, z)
    >>> print x
    ('new', -1, -2, -3)    
    >>> print y 
    ('new', -1, -2, -3)    
    >>> print org_x 
    ('org', 1, 2, 3)

    To reverse the process, do something like this:

    >>> z = pyjack.replace_all_refs(z, org_x)
    >>> del org_x
    >>> print x
    ('org', 1, 2, 3)
    >>> print y 
    ('org', 1, 2, 3)
    >>> print z
    ('new', -1, -2, -3)    

    .. note:
        The obj returned is, by the way, the last copy of :attr:`org_obj` in 
        memory; if you don't save a copy, there is no way to put state of the 
        system back to original state.     

    .. warning:: 

       This function does not work reliably on strings, due to how the 
       Python runtime interns strings. 

    """

    _gc.collect()

    hit = False
    for referrer in _gc.get_referrers(org_obj):

        # FRAMES -- PASS THEM UP
        if isinstance(referrer, _types.FrameType):
            continue

        # DICTS
        if isinstance(referrer, dict):

            cls = None

            # THIS CODE HERE IS TO DEAL WITH DICTPROXY TYPES
            if '__dict__' in referrer and '__weakref__' in referrer:
                for cls in _gc.get_referrers(referrer):
                    if _inspect.isclass(cls) and cls.__dict__ == referrer:
                        break

            for key, value in referrer.items():
                # REMEMBER TO REPLACE VALUES ...
                if value is org_obj:
                    hit = True
                    value = new_obj
                    referrer[key] = value
                    if cls:  # AGAIN, CLEANUP DICTPROXY PROBLEM
                        setattr(cls, key, new_obj)
                # AND KEYS.
                if key is org_obj:
                    hit = True
                    del referrer[key]
                    referrer[new_obj] = value

        # LISTS
        elif isinstance(referrer, list):
            for i, value in enumerate(referrer):
                if value is org_obj:
                    hit = True
                    referrer[i] = new_obj

        # SETS
        elif isinstance(referrer, set):
            referrer.remove(org_obj)
            referrer.add(new_obj)
            hit = True

        # TUPLE, FROZENSET
        elif isinstance(referrer, (tuple, frozenset,)):
            new_tuple = []
            for obj in referrer:
                if obj is org_obj:
                    new_tuple.append(new_obj)
                else:
                    new_tuple.append(obj)
            replace_all_refs(referrer, type(referrer)(new_tuple))

        # CELLTYPE        
        elif isinstance(referrer, _CELLTYPE):
            def proxy0(data):
                def proxy1(): return data

                return proxy1

            proxy = proxy0(new_obj)
            newcell = proxy.func_closure[0]
            replace_all_refs(referrer, newcell)

            # FUNCTIONS
        elif isinstance(referrer, _types.FunctionType):
            localsmap = {}
            for key in ['func_code', 'func_globals', 'func_name',
                        'func_defaults', 'func_closure']:
                orgattr = getattr(referrer, key)
                if orgattr is org_obj:
                    localsmap[key.split('func_')[-1]] = new_obj
                else:
                    localsmap[key.split('func_')[-1]] = orgattr
            localsmap['argdefs'] = localsmap['defaults']
            del localsmap['defaults']
            newfn = _types.FunctionType(**localsmap)
            replace_all_refs(referrer, newfn)

        # OTHER (IN DEBUG, SEE WHAT IS NOT SUPPORTED). 
        else:
            # debug: 
            # print type(referrer)
            pass

    if hit is False:
        raise AttributeError("Object '%r' not found" % org_obj)

    return org_obj


_func_code_map = {}


def _get_self():
    global _func_code_map

    frame = _inspect.currentframe()
    code = frame.f_back.f_code
    return _func_code_map[code]


class _PyjackFunc(object): pass


class _PyjackFuncCode(_PyjackFunc):

    def __init__(self, fn, proxyfn):
        global _func_code_map

        self._fn, self._proxyfn = fn, proxyfn

        def proxy(*args, **kwargs):
            import pyjack
            self = pyjack._get_self()
            return self._process_fn(args, kwargs)

        _func_code_map[proxy.func_code] = self
        self._org_func_code = fn.func_code
        self._proxy_func_code = proxy.func_code
        fn.func_code = proxy.func_code

        fn.restore = self.restore

    def _process_fn(self, args, kwargs):
        self._fn.func_code = self._org_func_code
        value = self._proxyfn(self._fn, *args, **kwargs)
        self._fn.func_code = self._proxy_func_code
        return value

    def restore(self):
        self._fn.func_code = self._org_func_code


class _PyjackFuncBuiltin(_PyjackFunc):

    def __init__(self, fn, proxyfn):
        self._fn = replace_all_refs(fn, self)
        self._proxyfn = proxyfn

    def __call__(self, *args, **kwargs):
        return self._proxyfn(self._fn, *args, **kwargs)

    def __getattr__(self, attr):
        try:
            return getattr(self._fn, attr)
        except AttributeError:
            bundle = (self._fn, attr,)
            raise AttributeError("function %r has no attr '%s'" % bundle)

    def restore(self):
        replace_all_refs(self, self._fn)


if __name__ == '__main__':
    import doctest

    doctest.testmod(optionflags=524)




from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import date
from datetime import datetime
from decimal import DecimalException
from enum import Enum

from dateutil import parser
from decimal import Decimal
from typing import Tuple
from typing import cast
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

from pytyped.macros.boxed import Boxed
from pytyped.macros.extractor import Extractor
from pytyped.macros.extractor import WithDefault
from pytyped.json.common import JsValue


class JsDecodeError(metaclass=ABCMeta):
    @abstractmethod
    def get_message(self) -> str:
        pass

    def to_string(self) -> str:
        return "Error when decoding JSON: " + self.get_message()


@dataclass
class JsDecodeErrorFinal(JsDecodeError):
    message: str

    def get_message(self) -> str:
        return ": " + self.message


@dataclass
class JsDecodeErrorInField(JsDecodeError):
    field_name: str
    error: JsDecodeError

    def get_message(self) -> str:
        return "/" + self.field_name + self.error.get_message()


@dataclass
class JsDecodeErrorInArray(JsDecodeError):
    index: int
    error: JsDecodeError

    def get_message(self) -> str:
        return "[" + str(self.index) + "]" + self.error.get_message()


class JsDecodeException(Exception):
    errors: List[JsDecodeError]

    def __init__(self, errors: List[JsDecodeError]) -> None:
        self.errors = errors

    def __str__(self) -> str:
        return "Found %d errors while validating JSON: [\n  %s]" % (
            len(self.errors),
            ",\n  ".join([e.to_string() for e in self.errors]),
        )


T = TypeVar("T")
TOrError = Union[Boxed[T], List[JsDecodeError]]


class JsonDecoder(Generic[T], metaclass=ABCMeta):
    is_optional: bool = False

    @abstractmethod
    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[T]:
        pass

    def read(self, json: JsValue) -> T:
        t_or_error = self.decode(json, [])
        if isinstance(t_or_error, Boxed):
            return t_or_error.t
        else:
            raise JsDecodeException(t_or_error)


@dataclass
class JsonObjectDecoder(JsonDecoder[T]):
    field_decoders: Dict[str, JsonDecoder[Any]]
    field_defaults: Dict[str, Union[Boxed[Any], Callable[[], Any]]]
    constructor: Callable[[Dict[str, Any]], T]

    def get_constructor(self) -> Callable[[Dict[str, Any]], T]:
        return self.constructor  # type: ignore

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[T]:
        if not isinstance(json, dict):
            return [
                JsDecodeErrorFinal(
                    "Expected a JSON object but received something else."
                )
            ]
        decoded_fields: Dict[str, Any] = {}
        decoding_errors: List[JsDecodeError] = []
        for field_name, field_decoder in self.field_decoders.items():
            if (field_name not in json) or (json[field_name] is None):
                if isinstance(field_decoder, JsonOptionalDecoder):
                    decoded_fields[field_name] = None
                elif field_name in self.field_defaults:
                    def_or_fac = self.field_defaults[field_name]
                    decoded_fields[field_name] = def_or_fac.t if isinstance(def_or_fac, Boxed) else def_or_fac()
                else:
                    decoding_errors.append(
                        JsDecodeErrorInField(
                            field_name,
                            JsDecodeErrorFinal(
                                "Non-optional field was not found"
                            ),
                        )
                    )
            else:
                # The field_name is a key and its associated value is not None. So, it should be decoded.
                field_json = cast(JsValue, json[field_name])
                decoded_field = field_decoder.decode(
                    field_json, [cast(JsValue, json)] + ancestors
                )
                if isinstance(decoded_field, Boxed):
                    decoded_fields[field_name] = decoded_field.t
                else:
                    for err in decoded_field:
                        decoding_errors.append(
                            JsDecodeErrorInField(field_name, err)
                        )

        if len(decoding_errors) > 0:
            return decoding_errors
        else:
            return Boxed(self.get_constructor()(decoded_fields))


@dataclass
class JsonTupleDecoder(JsonDecoder[Tuple[Any, ...]]):
    field_decoders: List[JsonDecoder[Any]]

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[Tuple[Any, ...]]:
        if not isinstance(json, list):
            return [
                JsDecodeErrorFinal(
                    "Expected a JSON array but received something else."
                )
            ]
        if len(json) != len(self.field_decoders):
            return [
                JsDecodeErrorFinal(
                    "Expected a JSON array of size %d but received one of size %d." % (len(self.field_decoders), len(json))
                )
            ]

        decoded_fields: List[Any] = []
        decoding_errors: List[JsDecodeError] = []
        for index, field_decoder in enumerate(self.field_decoders):
            # The field_name is a key and its associated value is not None. So, it should be decoded.
            field_json = cast(JsValue, json[index])
            decoded_field = field_decoder.decode(field_json, [cast(JsValue, json)] + ancestors)
            if isinstance(decoded_field, Boxed):
                decoded_fields.append(decoded_field.t)
            else:
                for err in decoded_field:
                    decoding_errors.append(
                        JsDecodeErrorInArray(index, err)
                    )

        if len(decoding_errors) > 0:
            return decoding_errors
        else:
            return Boxed(tuple(decoded_fields))


@dataclass
class JsonTaggedDecoder(JsonDecoder[Any]):
    branch_decoders: Dict[str, JsonDecoder[Any]]
    tag_field_name: str = "tag"
    value_field_name: Optional[str] = None  # None means the value is in the same object file

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[T]:
        if not isinstance(json, dict):
            return [JsDecodeErrorFinal("Expected a JSON object but received something else.")]

        tag_value = json.get(self.tag_field_name)
        if tag_value is None:
            return [
                JsDecodeErrorInField(
                    field_name=self.tag_field_name,
                    error=JsDecodeErrorFinal(
                        "Required tag field not found."
                    )
                )
            ]
        if not isinstance(tag_value, str):
            return [
                JsDecodeErrorInField(
                    field_name=self.tag_field_name,
                    error=JsDecodeErrorFinal(
                        "Expected a Json string for tag field but received something else."
                    )
                )
            ]

        decoder = self.branch_decoders.get(tag_value)
        if decoder is None:
            return [
                JsDecodeErrorInField(
                    field_name=self.tag_field_name,
                    error=JsDecodeErrorFinal(
                        "Unknown tag value %s (possible values are: %s)." % (tag_value, ", ".join(self.branch_decoders.keys()))
                    )
                )
            ]

        if self.value_field_name is None:
            return decoder.decode(json, ancestors)

        value_field = cast(JsValue, json.get(self.value_field_name))
        return decoder.decode(value_field, [value_field] + ancestors)


@dataclass
class JsonOptionalDecoder(JsonDecoder[Optional[T]]):
    inner_decoder: JsonDecoder[T]

    def decode(
        self, json: JsValue, ancestors: List[JsValue]
    ) -> TOrError[Optional[T]]:
        if json is None:
            return Boxed(None)

        result = self.inner_decoder.decode(json, ancestors)
        if isinstance(result, Boxed):
            return Boxed(result.t)
        else:
            return result


@dataclass
class JsonErrorAsDefaultDecoder(JsonDecoder[T]):
    inner_decoder: JsonDecoder[T]
    default: T

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[T]:
        result = self.inner_decoder.decode(json, ancestors)
        if isinstance(result, Boxed):
            return result
        return Boxed(self.default)


@dataclass
class JsonPriorityDecoder(JsonDecoder[Any]):
    # Decoders in the order of priority. The head of the list has more priority.
    inner_decoders: List[JsonDecoder[Any]]

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[Any]:
        errors: List[JsDecodeError] = []
        for inner_decoder in self.inner_decoders:
            result = inner_decoder.decode(json, ancestors)
            if isinstance(result, Boxed):
                return result
            for e in result:
                errors.append(e)

        return errors


U = TypeVar("U")


@dataclass
class JsonMappedDecoder(JsonDecoder[T], Generic[T, U]):
    u_decoder: JsonDecoder[U]
    u_to_t: Callable[[U], T]

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[T]:
        result = self.u_decoder.decode(json, ancestors)
        if isinstance(result, Boxed):
            transformer = cast(Callable[[U], T], self.u_to_t)
            return Boxed(transformer(result.t))
        return result


@dataclass
class JsonFlatMappedDecoder(JsonDecoder[T], Generic[T, U]):
    u_decoder: JsonDecoder[U]
    u_to_t: Callable[[U], TOrError[T]]

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[T]:
        result = self.u_decoder.decode(json, ancestors)
        if isinstance(result, Boxed):
            transformer = cast(Callable[[U], TOrError[T]], self.u_to_t)
            return transformer(result.t)
        return result


@dataclass
class JsonBoxedDecoder(JsonDecoder[T]):
    field_name: str
    field_decoder: JsonDecoder[Any]
    constructor: Callable[[Dict[str, Any]], T]

    def get_constructor(self) -> Callable[[Dict[str, Any]], T]:
        return self.constructor  # type: ignore

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[T]:
        decoded_field = self.field_decoder.decode(json, ancestors)
        if isinstance(decoded_field, Boxed):
            return Boxed(
                self.get_constructor()({self.field_name: decoded_field.t})
            )
        else:
            return decoded_field


@dataclass
class JsonListDecoder(JsonDecoder[List[T]]):
    element_decoder: JsonDecoder[T]

    def decode(
        self, json: JsValue, ancestors: List[JsValue]
    ) -> TOrError[List[T]]:
        if not isinstance(json, list):
            return [
                JsDecodeErrorFinal(
                    "Expected a JSON array but received something else."
                )
            ]
        decoded_elements: List[Any] = []
        decoding_errors: List[JsDecodeError] = []
        for index, element_json in enumerate(json):
            element_json = cast(JsValue, element_json)
            parent: JsValue = json
            decoded_element = self.element_decoder.decode(
                element_json, [parent] + ancestors
            )
            if isinstance(decoded_element, Boxed):
                decoded_elements.append(decoded_element.t)
            else:
                for err in decoded_element:
                    decoding_errors.append(JsDecodeErrorInArray(index, err))

        if len(decoding_errors) > 0:
            return decoding_errors
        else:
            return Boxed(decoded_elements)


@dataclass
class JsonStringDictionaryDecoder(JsonDecoder[Dict[str, T]]):
    element_decoder: JsonDecoder[T]

    def decode(
        self, json: JsValue, ancestors: List[JsValue]
    ) -> TOrError[Dict[str, T]]:
        if not isinstance(json, dict):
            return [
                JsDecodeErrorFinal(
                    "Expected a JSON object but received something else."
                )
            ]
        decoded_elements: Dict[str, Any] = {}
        decoding_errors: List[JsDecodeError] = []
        for key, value_json in json.items():
            if type(key) is not str:
                decoding_errors.append(
                    JsDecodeErrorFinal("Found non-string key %s in a json object (type: %s)." % (str(key), type(key)))
                )
            else:
                value_json = cast(JsValue, value_json)
                parent: JsValue = json
                decoded_value = self.element_decoder.decode(
                    value_json, [parent] + ancestors
                )
                if isinstance(decoded_value, Boxed):
                    if key not in decoded_elements:
                        decoded_elements[key] = decoded_value.t
                    else:
                        decoding_errors.append(JsDecodeErrorFinal("Found key %s more than once." % key))
                else:
                    for err in decoded_value:
                        decoding_errors.append(JsDecodeErrorInField(key, err))

        if len(decoding_errors) > 0:
            return decoding_errors
        else:
            return Boxed(decoded_elements)


@dataclass
class JsonStringDecoder(JsonDecoder[str]):
    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[str]:
        if not isinstance(json, str):
            return [
                JsDecodeErrorFinal(
                    "Expected a JSON string but received something else."
                )
            ]
        return Boxed(json)


@dataclass
class JsonNumberDecoder(JsonDecoder[Decimal]):
    def decode(
        self, json: JsValue, ancestors: List[JsValue]
    ) -> TOrError[Decimal]:
        if isinstance(json, Decimal):
            return Boxed(json)
        elif (
            isinstance(json, float)
            or isinstance(json, int)
            or isinstance(json, str)
        ):
            try:
                return Boxed(Decimal(json))
            except DecimalException:
                return [JsDecodeErrorFinal("Value not convertible to decimal: '%s'." % str(json))]

        return [
            JsDecodeErrorFinal(
                "Expected a JSON number or a JSON string encoding a number but received something of type %s."
                % type(json)
            )
        ]


@dataclass
class JsonBooleanDecoder(JsonDecoder[bool]):
    def decode(
        self, json: JsValue, ancestors: List[JsValue]
    ) -> TOrError[bool]:
        if not isinstance(json, bool):
            return [
                JsDecodeErrorFinal(
                    "Expected a JSON boolean but received something else."
                )
            ]
        return Boxed(json)


json_string_decoder = JsonStringDecoder()
json_number_decoder = JsonNumberDecoder()


@dataclass
class JsonIntegerDecoder(JsonDecoder[int]):
    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[int]:
        decimal_or_error = json_number_decoder.decode(json, ancestors)
        if isinstance(decimal_or_error, Boxed):
            d = decimal_or_error.t
            if int(d) == d:
                return Boxed(int(d))
            else:
                return [
                    JsDecodeErrorFinal(
                        "Expected an integral number but received non-intgeral number."
                    )
                ]
        else:
            return decimal_or_error


@dataclass
class JsonDateDecoder(JsonDecoder[date]):
    def decode(
        self, json: JsValue, ancestors: List[JsValue]
    ) -> TOrError[date]:
        string_or_error = json_string_decoder.decode(json, ancestors)
        if isinstance(string_or_error, Boxed):
            s = string_or_error.t
            try:
                return Boxed(parser.isoparse(s).date())
            except ValueError:
                return [
                    JsDecodeErrorFinal(
                        "Expected a string representing a date but received '%s'."
                        % s
                    )
                ]
        else:
            return string_or_error


@dataclass
class JsonDatetimeDecoder(JsonDecoder[datetime]):
    def decode(
        self, json: JsValue, ancestors: List[JsValue]
    ) -> TOrError[datetime]:
        string_or_error = json_string_decoder.decode(json, ancestors)
        if isinstance(string_or_error, Boxed):
            s = string_or_error.t
            try:
                return Boxed(parser.isoparse(s))
            except ValueError:
                return [
                    JsDecodeErrorFinal(
                        "Expected a string representing a datetime but received '%s'."
                        % s
                    )
                ]
        else:
            return string_or_error


@dataclass
class JsonEnumDecoder(JsonDecoder[T]):
    enum_name: str
    enum_values: Dict[str, T]

    def decode(self, json: JsValue, ancestors: List[JsValue]) -> TOrError[T]:
        string_or_error = json_string_decoder.decode(json, ancestors)
        if isinstance(string_or_error, Boxed):
            s = string_or_error.t
            enum_value = self.enum_values.get(s)
            if enum_value is None:
                return [
                    JsDecodeErrorFinal(
                        "Unexpected value %s while deserializing enum %s."
                        % (s, self.enum_name)
                    )
                ]
            else:
                return Boxed(enum_value)
        else:
            return string_or_error


@dataclass
class JsonNoneDecoder(JsonDecoder[None]):
    def decode(
        self, json: JsValue, ancestors: List[JsValue]
    ) -> TOrError[None]:
        return Boxed(None)


class AutoJsonDecoder(Extractor[JsonDecoder[Any]]):
    json_boolean_decoder = JsonBooleanDecoder()
    json_integer_decoder = JsonIntegerDecoder()
    json_date_decoder = JsonDateDecoder()
    json_datetime_decoder = JsonDatetimeDecoder()

    basic_json_decoders: Dict[type, Boxed[JsonDecoder[Any]]] = {
        bool: Boxed(json_boolean_decoder),
        str: Boxed(json_string_decoder),
        int: Boxed(json_integer_decoder),
        Decimal: Boxed(json_number_decoder),
        datetime: Boxed(json_datetime_decoder),
        date: Boxed(json_date_decoder),
        type(None): Boxed(cast(JsonDecoder[Any], JsonNoneDecoder))
    }

    def __init__(self) -> None:
        super().__init__()

    @property
    def basics(self) -> Dict[type, Boxed[JsonDecoder[Any]]]:
        return self.basic_json_decoders

    def named_product_extractor(self, t: type, fields: Dict[str, WithDefault[JsonDecoder[Any]]]) -> JsonDecoder[Any]:
        field_decoders: Dict[str, JsonDecoder[Any]] = {
            n: v.t for (n, v) in fields.items()
        }
        field_defaults: Dict[str, Union[Boxed[Any], Callable[[], Any]]] = {
            n: v.default for (n, v) in fields.items() if v.default is not None
        }

        return JsonObjectDecoder(
            field_decoders=field_decoders,
            field_defaults=field_defaults,
            constructor=lambda args: t(**args),
        )

    def unnamed_product_extractor(self, t: type, fields: List[JsonDecoder[Any]]) -> JsonDecoder[Tuple[Any, ...]]:
        return JsonTupleDecoder(fields)

    def named_sum_extractor(self, t: type, branches: Dict[str, Tuple[type, JsonDecoder[Any]]]) -> JsonDecoder[Any]:
        return JsonTaggedDecoder(
            branch_decoders={s: t for (s, (_, t)) in branches.items()},
            tag_field_name=t.__name__, value_field_name=None
        )

    def unnamed_sum_extractor(self, t: type, branches: List[Tuple[type, JsonDecoder[Any]]]) -> JsonDecoder[Any]:
        return JsonPriorityDecoder([d for (_, d) in branches])

    def optional_extractor(self, t: JsonDecoder[T]) -> JsonDecoder[Optional[T]]:
        return JsonOptionalDecoder(t)

    def list_extractor(self, t: JsonDecoder[T]) -> JsonDecoder[List[T]]:
        return JsonListDecoder(t)

    def dictionary_extractor(
        self,
        key: type,
        value: type,
        key_ext: JsonDecoder[Any],
        val_ext: JsonDecoder[Any]
    ) -> JsonDecoder[Dict[Any, Any]]:
        def to_enum_dict(d: Dict[str, Any]) -> TOrError[Dict[Any, Any]]:
            result: Dict[Any, Any] = {}
            errors: List[JsDecodeError] = []
            for k, v in d.items():
                try:
                    e = cast(Enum, key_ext.read(k))
                    result[e] = v
                except JsDecodeException as exception:
                    for err in exception.errors:
                        errors.append(err)

            if len(errors) > 0:
                return errors
            return Boxed(result)

        if key is str:
            return JsonStringDictionaryDecoder(val_ext)
        if issubclass(key, Enum):
            return JsonFlatMappedDecoder(JsonStringDictionaryDecoder(val_ext), to_enum_dict)
        raise NotImplementedError()

    def enum_extractor(self, enum_name: str, enum_values: List[Tuple[str, Any]]) -> JsonDecoder[Any]:
        return JsonEnumDecoder(enum_name, {n: v for (n, v) in enum_values})



from abc import ABCMeta
from abc import abstractmethod
from dataclasses import Field
from dataclasses import MISSING, dataclass, is_dataclass
from enum import Enum
from inspect import isclass
from typing import FrozenSet
from typing import cast
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

from pytyped.macros.boxed import Boxed
from pytyped.macros.pyjack import replace_all_refs


@dataclass
class UnknownExtractorException(Exception):
    t: type

    def message(self) -> str:
        return "Automatic extraction not implemented for type %s" % str(self.t)


@dataclass
class ExtractorAssignmentException(Exception):
    assignments: Dict[str, type]
    t: str

    def message(self) -> str:
        return "Automatic extraction not implemented for type %s" % str(self.t)


T = TypeVar("T")
U = TypeVar("U")


@dataclass
class WithDefault(Generic[T]):
    t: T
    default: Union[None, Boxed[Any], Callable[[], Any]]

    def map(self, f: Callable[[T], U]) -> "WithDefault[U]":
        return WithDefault(f(self.t), self.default)


FieldType = WithDefault[type]


@dataclass
class UnnamedUnionDescriptor:
    branches: List[type]  # List of non-None branches
    is_optional: bool  # True if `None` is one of the union branches


@dataclass
class NamedUnionDescriptor:
    branches: Dict[str, type]  # Mapping of names to branches


@dataclass
class RecursiveTypeApplication:
    ref_count: int
    typ: type
    assignments: FrozenSet[Tuple[str, type]]

class Extractor(Generic[T], metaclass=ABCMeta):
    # Memoized values for types that have already been extracted.
    # Since types can be generic, their relative context are included.
    # So, the mapping for a generic type G[X] would be (G, {X --> A}) --> T[G[A]].
    memoized: Dict[Tuple[type, FrozenSet[Tuple[str, type]]], Boxed[T]]

    # Current context: A mapping from type variable names to their types.
    # Should only be non-empty when a type-extraction is in progress.
    # Not thread-safe.
    _context: Dict[str, type]

    def __init__(self) -> None:
        self.memoized = {}
        self._context = {}

    @property
    @abstractmethod
    def basics(self) -> Dict[type, Boxed[T]]:
        # Mapping from X -> T[X] for basic types (usually str, int, bool, float, etc)
        pass

    @abstractmethod
    def named_product_extractor(self, t: type, fields: Dict[str, WithDefault[T]]) -> T:
        # Given N = {f1: X1, f2: X2, ..., fn: Xn}, T[X1], T[X2], ..., and T[Xn], generates T[N]
        pass

    @abstractmethod
    def unnamed_product_extractor(self, t: type, fields: List[T]) -> T:
        # Given N = {f1: X1, f2: X2, ..., fn: Xn}, T[X1], T[X2], ..., and T[Xn], generates T[N]
        pass

    @abstractmethod
    def named_sum_extractor(self, t: type, branches: Dict[str, Tuple[type, T]]) -> T:
        # Given T[X], generates T[Optional[X]]
        pass

    @abstractmethod
    def unnamed_sum_extractor(self, t: type, branches: List[Tuple[type, T]]) -> T:
        # Given T[X], generates T[Optional[X]]
        pass

    @abstractmethod
    def optional_extractor(self, t: T) -> T:
        # Given T[X], generates T[Optional[X]]
        pass

    @abstractmethod
    def list_extractor(self, t: T) -> T:
        # Given T[X], generates T[List[X]]
        pass

    @abstractmethod
    def dictionary_extractor(self, key: type, value: type, key_ext: T, val_ext: T) -> T:
        # Given T[X], generates T[List[X]]
        pass

    @abstractmethod
    def enum_extractor(self, enum_name: str, enum_values: List[Tuple[str, Any]]) -> T:
        # Given enum_name: str, and enum_values: Dict[str, E], generates T[E]
        pass

    @staticmethod
    def extract_if_union_type(t: type) -> Optional[UnnamedUnionDescriptor]:
        """
        :param t: Type that needs to be checked for being optional.
        :return: Returns Boxed(x) if t is Optional[x] and returns None otherwise.
        """
        if not hasattr(t, "__origin__"):
            return None
        if t.__origin__ is not Union:  # type: ignore
            return None
        t = cast(Union, t)
        args: Tuple[type, ...] = cast(Tuple[type, ...], t.__args__)  # type: ignore

        is_optional: bool = False
        branches: List[type] = []
        for inner_type in args:
            if inner_type is type(None):  # noqa: E721
                is_optional = True
            else:
                branches.append(inner_type)

        return UnnamedUnionDescriptor(branches, is_optional)

    @staticmethod
    def extract_if_has_sub_classes(t: type) -> Optional[NamedUnionDescriptor]:
        """
        :param t: Type that needs to be checked for being optional.
        :return: Returns Boxed(x) if t is Optional[x] and returns None otherwise.
        """
        if not (isclass(t) or (hasattr(t, "__origin__") and hasattr(t.__origin__, "__args__") and isclass(t.__origin__))):
            return None

        branch_list: List[type]
        if hasattr(t, "__subclasses__"):
            branch_list = t.__subclasses__()
        elif hasattr(t, "__origin__") and hasattr(t.__origin__, "__subclasses__"):
            branch_list = t.__origin__.__subclasses__()
        else:
            return None
        if len(branch_list) <= 0:
            return None

        branches: Dict[str, type] = {}
        for branch in branch_list:
            if not hasattr(branch, "__name__"):
                return None
            branches[branch.__name__] = branch

        return NamedUnionDescriptor(branches)

    @staticmethod
    def extract_if_tuple_type(t: type) -> Optional[List[type]]:
        """
        :param t: Type that needs to be checked for being a named tuple.
        :return:
           If `t` is a Tuple, returns a list of `t`'s field types.
           Otherwise, returns None.
        """
        if not hasattr(t, "__origin__"):
            return None
        if t.__origin__ not in [Tuple, tuple]:  # type: ignore
            return None
        if not hasattr(t, "__args__"):
            return None

        args: Tuple[type, ...] = cast(Tuple[type, ...], t.__args__)  # type: ignore
        inner_types: List[type] = [t for t in args]
        return inner_types

    @staticmethod
    def extract_if_named_tuple_type(t: type) -> Optional[Dict[str, FieldType]]:
        """
        :param t: Type that needs to be checked for being a named tuple.
        :return:
           If `t` is a NamedTuple, returns a mapping from `t`'s field names to their types and default values.
           Otherwise, returns None.
        """
        if not hasattr(t, "_field_types"):
            return None

        t = cast(Type[NamedTuple], t)
        fields: Dict[str, FieldType] = {}
        for (f_name, f_type) in t._field_types.items():
            f_default: Optional[Boxed[Any]] = None
            if f_name in t._field_defaults:
                f_default = Boxed(t._field_defaults[f_name])
            fields[f_name] = FieldType(f_type, f_default)
        return fields

    @staticmethod
    def extract_if_dataclass_type(t: type) -> Optional[Dict[str, FieldType]]:
        """
        :param t: Type that needs to be checked for being a dataclass.
        :return:
           If `t` is a dataclass, returns a mapping from `t`'s field names to their types and default values.
           Otherwise, returns None.
        """
        dataclass_fields: Dict[str, Field]
        if is_dataclass(t):
            dataclass_fields = cast(Dict[str, Field], t.__dataclass_fields__)  # type: ignore
        elif hasattr(t, "__origin__") and hasattr(t.__origin__, "__args__") and is_dataclass(t.__origin__):
            dataclass_fields = cast(Dict[str, Field], t.__origin__.__dataclass_fields__)  # type: ignore
        else:
            return None

        fields: Dict[str, FieldType] = {}
        for (field_name, field_definition) in dataclass_fields.items():
            field_default: Union[None, Boxed[Any], Callable[[], Any]] = None
            if field_definition.default is not MISSING:
                field_default = Boxed(field_definition.default)
            elif field_definition.default_factory is not MISSING:  # type: ignore
                field_default = field_definition.default_factory  # type: ignore
            fields[field_name] = FieldType(field_definition.type, field_default)
        return fields

    @staticmethod
    def extract_if_list_type(t: type) -> Union[None, str, Boxed[type]]:
        """
        :param t: type that needs to be checked for being a list.
        :return: Returns `Boxed("X")` if `t` is the type `List[X]` with X being the type variable name used by `List`.
        """
        if not hasattr(t, "__origin__"):
            return None
        origin = cast(type, t.__origin__)  # type: ignore
        if origin is List:
            return origin.__parameters__[0].__name__  # type: ignore
        if origin is list:
            return Boxed(t.__args__[0])  # type: ignore
        return None

    @staticmethod
    def extract_if_dictionary_type(t: type) -> Union[None, Tuple[str, str], Boxed[Tuple[type, type]]]:
        """
        :param t: type that needs to be checked for being a dictionary.
        :return: Returns `Boxed(("X", "Y"))` if `t` is the type `Dict[X, Y]` and X and Y are type variable names
                 for key and value types respectively.
        """
        if not hasattr(t, "__origin__"):
            return None
        origin = cast(type, t.__origin__)  # type: ignore
        if origin is Dict:
            return origin.__parameters__[0].__name__, origin.__parameters__[1].__name__  # type: ignore
        if origin is dict:
            return Boxed((t.__args__[0], t.__args__[1]))  # type: ignore
        return None

    @staticmethod
    def or_else(t: Optional[T], f: Callable[[], Optional[T]]) -> Optional[T]:
        if t is None:
            return f()
        return t

    def assignments(self, t: type) -> Tuple[type, FrozenSet[Tuple[str, type]], Dict[str, type]]:
        new_context = self._context.copy()
        if hasattr(t, "__origin__") and hasattr(t.__origin__, "__parameters__") and hasattr(t, "__args__"):
            t_origin = t.__origin__
            assert len(t.__args__) == len(t_origin.__parameters__)
            for parameter, arg in zip(t_origin.__parameters__, t.__args__):  # type:ignore
                parameter_name: str = cast(str, parameter.__name__)
                if isinstance(arg, TypeVar):  # type: ignore
                    arg_name: str = arg.__name__
                    if arg_name not in self._context:
                        raise ExtractorAssignmentException(self._context, arg_name)
                    new_context[parameter_name] = self._context[arg_name]
                else:
                    if not hasattr(arg, "__parameters__") or len(arg.__parameters__) <= 0:
                        new_context[parameter_name] = arg
                    else:
                        concretized_params_list: List[type] = []
                        for p in arg.__parameters__:
                            p_name = p.__name__
                            p_type = self._context.get(p_name)
                            if p_type is None:
                                raise ExtractorAssignmentException(self._context, p_name)
                            concretized_params_list.append(p_type)
                        concretized_params_tuple = tuple(t for t in concretized_params_list)
                        new_context[parameter_name] = arg[concretized_params_tuple]
        elif not hasattr(t, "__origin__"):
            t_origin = t
        else:
            return t, frozenset(), self._context

        assignments: Dict[str, type] = {}
        if hasattr(t_origin, "__parameters__"):
            for parameter in t_origin.__parameters__:  # type:ignore
                parameter_name: str = cast(str, parameter.__name__)
                parameter_type = new_context.get(parameter_name)
                if parameter_type is None:
                    raise ExtractorAssignmentException(new_context, parameter_name)
                assignments[parameter_name] = parameter_type

        if len(assignments) <= 0:
            return t_origin, frozenset(), self._context

        return t_origin, frozenset(assignments.items()), new_context
 
    def add_special(self, typ: type, value: T) -> None:
        t_origin, t_assignment, _ = self.assignments(typ)
        self.memoized[(t_origin, t_assignment)] = Boxed(value)

    def _var_to_type(self, var_name: str) -> type:
        if var_name not in self._context:
            raise ExtractorAssignmentException(self._context.copy(), var_name)
        list_inner_type = self._context[var_name]
        return list_inner_type

    def _extract_basic_type(self, t: type) -> Optional[Boxed[T]]:
        return self.basics.get(t)

    def _extract_unnamed_product_type(self, product_type: type) -> Optional[Boxed[T]]:
        maybe_fields = Extractor.extract_if_tuple_type(product_type)
        if maybe_fields is None:
            return None

        fields: List[T] = [self._make(t) for t in maybe_fields]
        return Boxed(self.unnamed_product_extractor(product_type, fields))

    def _extract_named_product_type(self, product_type: type) -> Optional[Boxed[T]]:
        maybe_fields = Extractor.extract_if_named_tuple_type(product_type)
        maybe_fields = Extractor.or_else(maybe_fields, lambda: Extractor.extract_if_dataclass_type(product_type))
        if maybe_fields is None:
            return None

        fields: Dict[str, WithDefault[T]] = {
            n: v.map(self._make) for (n, v) in maybe_fields.items()
        }
        return Boxed(self.named_product_extractor(product_type, fields))

    def _extract_unnamed_sum_type(self, sum_type: type) -> Optional[Boxed[T]]:
        maybe_unnamed_sum_type = Extractor.extract_if_union_type(sum_type)
        if maybe_unnamed_sum_type is None:
            return None

        extracted_branches: List[Tuple[type, T]] = [(t, self._make(t)) for t in maybe_unnamed_sum_type.branches]
        extracted_union: T
        if len(extracted_branches) == 1:
            extracted_union = extracted_branches[0][1]
        else:
            extracted_union = self.unnamed_sum_extractor(sum_type, extracted_branches)
        if maybe_unnamed_sum_type.is_optional:
            extracted_union = self.optional_extractor(extracted_union)
        return Boxed(extracted_union)

    def _extract_named_sum_type(self, sum_type: type) -> Optional[Boxed[T]]:
        maybe_named_sum_type = Extractor.extract_if_has_sub_classes(sum_type)
        if maybe_named_sum_type is None:
            return None

        extracted_branches: Dict[str, Tuple[type, T]] = {
            s: (t, self._make(t)) for (s, t) in maybe_named_sum_type.branches.items()
        }
        return Boxed(self.named_sum_extractor(sum_type, extracted_branches))

    def _extract_list_type(self, list_type: type) -> Optional[Boxed[T]]:
        maybe_list_var_name = Extractor.extract_if_list_type(list_type)
        if maybe_list_var_name is None:
            return None

        list_inner_type: type
        if isinstance(maybe_list_var_name, str):
            list_inner_type = self._var_to_type(maybe_list_var_name)
        else:
            list_inner_type = maybe_list_var_name.t

        return Boxed(self.list_extractor(self._make(list_inner_type)))

    def _extract_dictionary_type(self, dict_type: type) -> Optional[Boxed[T]]:
        maybe_dict_vars = Extractor.extract_if_dictionary_type(dict_type)
        if maybe_dict_vars is None:
            return None

        key_type: type
        value_type: type
        if isinstance(maybe_dict_vars, Boxed):
            key_type, value_type = maybe_dict_vars.t
        else:
            (key_var_name, value_var_name) = maybe_dict_vars
            key_type = self._var_to_type(key_var_name)
            value_type = self._var_to_type(value_var_name)

        key_extractor = self._make(key_type)
        value_extractor = self._make(value_type)
        return Boxed(self.dictionary_extractor(key_type, value_type, key_extractor, value_extractor))

    def _extract_enum_type(self, enum_type: type) -> Optional[Boxed[T]]:
        try:
            if not issubclass(enum_type, Enum):
                return None
        except Exception:
            return None

        value_dict = cast(Dict[str, Any], enum_type._value2member_map_)
        value_list = cast(List[Tuple[str, Any]], list(value_dict.items()))
        return Boxed(self.enum_extractor(str(enum_type), value_list))

    def _make(self, t: type) -> T:
        if isinstance(t, TypeVar):  # type: ignore
            t = self._var_to_type(t.__name__)

        t_origin, t_assignments, new_context = self.assignments(t)
        key = (t_origin, t_assignments)
        if key in self.memoized:
            result = self.memoized[key].t
            if isinstance(result, RecursiveTypeApplication):
                result.ref_count = result.ref_count + 1
            return result

        recursion_placeholder = RecursiveTypeApplication(0, t_origin, t_assignments)
        self.memoized[key] = Boxed(recursion_placeholder)

        old_context = self._context
        self._context = new_context

        result: Optional[Boxed[T]] = None
        result = Extractor.or_else(result, lambda: self._extract_basic_type(t_origin))
        result = Extractor.or_else(result, lambda: self._extract_unnamed_sum_type(t_origin))
        result = Extractor.or_else(result, lambda: self._extract_named_sum_type(t_origin))
        result = Extractor.or_else(result, lambda: self._extract_list_type(t_origin))
        result = Extractor.or_else(result, lambda: self._extract_dictionary_type(t_origin))
        result = Extractor.or_else(result, lambda: self._extract_unnamed_product_type(t_origin))
        result = Extractor.or_else(result, lambda: self._extract_named_product_type(t_origin))
        result = Extractor.or_else(result, lambda: self._extract_enum_type(t_origin))

        if result is None:
            raise UnknownExtractorException(t_origin)

        if recursion_placeholder.ref_count > 0:
            replace_all_refs(recursion_placeholder, result.t)
        self._context = old_context
        self.memoized[key] = result
        return result.t

    def extract(self, in_typ: type) -> T:
        """
        :param in_typ: The type for which meta-programming needs to be initiated (i.e., generate T[in_typ])
        :param assumptions: A mapping from type variable names to their assumed generated types. This is useful when
            extracting the type of a generic class without knowing the type of its variables.
        :return: The auto generated value T[in_typ]
        """
        self._context = {}
        result: T = self._make(t=in_typ)
        assert len(self._context) == 0, "Non-empty context at the top level."

        return result



from abc import ABCMeta
from abc import abstractmethod
from dataclasses import dataclass
from datetime import date
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

from pytyped.macros.boxed import Boxed
from pytyped.macros.extractor import Extractor
from pytyped.macros.extractor import WithDefault
from pytyped.metrics.common import MetricsBranch
from pytyped.metrics.common import MetricsLeaf
from pytyped.metrics.common import MetricsNone
from pytyped.metrics.common import MetricsTags
from pytyped.metrics.common import MetricsTree

T = TypeVar("T")


class MetricsExporter(Generic[T], metaclass=ABCMeta):
    @abstractmethod
    def outer_tags(self, names: List[str], t: T) -> Dict[str, str]:
       pass

    @abstractmethod
    def export(self, names: List[str], t: T) -> MetricsTree:
        pass


@dataclass
class ValueExporter(MetricsExporter[Union[int, float, Decimal]]):
    def outer_tags(self, names: List[str], t: Union[int, float, Decimal]) -> Dict[str, str]:
        return {}

    def export(self, names: List[str], t: Union[int, float, Decimal]) -> MetricsTree:
        return MetricsLeaf(".".join(names), t)


@dataclass
class TagExporter(MetricsExporter[str]):
    def outer_tags(self, names: List[str], t: str) -> Dict[str, str]:
        return {".".join(names): t}

    def export(self, names: List[str], t: str) -> MetricsTree:
        return MetricsNone()


@dataclass
class BooleanExporter(MetricsExporter[bool]):
    def outer_tags(self, names: List[str], t: bool) -> Dict[str, str]:
        return {".".join(names): "yes" if t else "no"}

    def export(self, names: List[str], t: bool) -> MetricsTree:
        return MetricsNone()


class DateExporter(MetricsExporter[date]):
    _weekday_names: List[str] = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]

    def outer_tags(self, names: List[str], t: date) -> Dict[str, str]:
        name = ".".join(names)
        return {
            name + "_day": str(t.day),
            name + "_month": str(t.month),
            name + "_year": str(t.year),
            name + "_weekday": DateExporter._weekday_names[t.weekday()]
        }

    def export(self, names: List[str], t: T) -> MetricsTree:
        return MetricsNone()


class DateTimeExporter(MetricsExporter[datetime]):
    _date_exporter: DateExporter = DateExporter()

    def outer_tags(self, names: List[str], t: datetime) -> Dict[str, str]:
        tags = self._date_exporter.outer_tags(names, t.date())
        tags[".".join(names) + "_hour"] = str(t.hour)

        return tags

    def export(self, names: List[str], t: datetime) -> MetricsTree:
        return MetricsNone()


@dataclass
class NamedProductExporter(MetricsExporter[T]):
    field_exporters: Dict[str, MetricsExporter[Any]]

    def outer_tags(self, names: List[str], t: T) -> Dict[str, str]:
        return {}

    def export(self, names: List[str], t: T) -> MetricsTree:
        d: Dict[str, Any]
        if hasattr(t, "_asdict"):
            d = cast(NamedTuple, t)._asdict()
        else:
            d = t.__dict__

        tags: Dict[str, str] = {}
        for field_name, field_exporter in self.field_exporters.items():
            tags.update(field_exporter.outer_tags(names + [field_name], d[field_name]))

        result: MetricsTree = MetricsBranch([
            field_exporter.export(names + [field_name], d[field_name])
            for field_name, field_exporter in self.field_exporters.items()
        ])
        if len(tags) > 0:
            result = MetricsTags(tags, result)

        return result


@dataclass
class TupleExporter(MetricsExporter[Tuple[Any, ...]]):
    exporters: List[MetricsExporter[Any]]

    def outer_tags(self, names: List[str], t: Tuple[Any, ...]) -> Dict[str, str]:
        return {}

    def export(self, names: List[str], t: Tuple[Any, ...]) -> MetricsTree:
        return MetricsBranch([e.export(names, v) for v, e in zip(t, self.exporters)])


@dataclass
class ListExporter(MetricsExporter[List[T]]):
    inner_exporter: MetricsExporter[T]

    def outer_tags(self, names: List[str], t: List[T]) -> Dict[str, str]:
        return {}

    def export(self, names: List[str], t: List[T]) -> MetricsTree:
        return MetricsBranch([self.inner_exporter.export(names, v) for v in t])


@dataclass
class TaggedExporter(MetricsExporter[T]):
    branches: Dict[str, Tuple[type, MetricsExporter[Any]]]
    tag_tag: str

    def outer_tags(self, names: List[str], t: T) -> Dict[str, str]:
        return {}

    def export(self, names: List[str], t: T) -> MetricsTree:
        for branch_name, (branch_type, exporter) in self.branches.items():
            if isinstance(t, branch_type):
                exported_value: MetricsTree = exporter.export(names, t)
                return MetricsTags({".".join(names + [self.tag_tag]): branch_name}, exported_value)

        return MetricsNone()


@dataclass
class PriorityExporter(MetricsExporter[T]):
    branches: List[Tuple[type, MetricsExporter[Any]]]

    def outer_tags(self, names: List[str], t: T) -> Dict[str, str]:
        return {}

    def export(self, names: List[str], t: T) -> MetricsTree:
        for branch_type, exporter in self.branches:
            if isinstance(t, branch_type):
                return exporter.export(names, t)

        return MetricsNone()


@dataclass
class OptionalExporter(MetricsExporter[Optional[T]]):
    exporter: MetricsExporter[T]

    def outer_tags(self, names: List[str], t: Optional[T]) -> Dict[str, str]:
        return {".".join(names) + "_present": "no" if t is None else "yes"}

    def export(self, names: List[str], t: Optional[T]) -> MetricsTree:
        return MetricsNone() if t is None else self.exporter.export(names, t)


@dataclass
class StringDictionaryExporter(MetricsExporter[Dict[str, T]]):
    element_exporter: MetricsExporter[T]

    def outer_tags(self, names: List[str], t: Dict[str, T]) -> Dict[str, str]:
        return {}

    def export(self, names: List[str], d: Dict[str, T]) -> MetricsTree:
        name = ".".join(names)
        prev_names = names[:-1]
        return MetricsBranch([MetricsTags({name: k}, self.element_exporter.export(prev_names, v)) for k, v in d.items()])


@dataclass
class EnumExporter(MetricsExporter[Enum]):
    def outer_tags(self, names: List[str], t: Enum) -> Dict[str, str]:
        return {".".join(names): str(t.value)}

    def export(self, names: List[str], t: Enum) -> MetricsTree:
        return MetricsNone()


class AutoMetricExporter(Extractor[MetricsExporter[Any]]):
    _enum_exporter = EnumExporter()
    _basics: Dict[type, Boxed[MetricsExporter[Any]]] = {
        bool: Boxed(BooleanExporter()),
        int: Boxed(ValueExporter()),
        float: Boxed(ValueExporter()),
        Decimal: Boxed(ValueExporter()),
        str: Boxed(TagExporter()),
        date: Boxed(DateExporter()),
        datetime: Boxed(DateTimeExporter())
    }

    @property
    def basics(self) -> Dict[type, Boxed[MetricsExporter[Any]]]:
        return self._basics

    def named_product_extractor(
        self,
        t: type,
        fields: Dict[str, WithDefault[MetricsExporter[Any]]]
    ) -> MetricsExporter[Any]:
        return NamedProductExporter(field_exporters={f_name: f_exporter.t for f_name, f_exporter in fields.items()})

    def unnamed_product_extractor(self, t: type, fields: List[MetricsExporter[Any]]) -> MetricsExporter[Any]:
        return TupleExporter(fields)

    def named_sum_extractor(
        self,
        t: type,
        branches: Dict[str, Tuple[type, MetricsExporter[Any]]]
    ) -> MetricsExporter[Any]:
        return TaggedExporter(branches=branches, tag_tag=t.__name__)

    def unnamed_sum_extractor(self, t: type, branches: List[Tuple[type, MetricsExporter[Any]]]) -> MetricsExporter[Any]:
        return PriorityExporter(branches=branches)

    def optional_extractor(self, t: MetricsExporter[Any]) -> MetricsExporter[Any]:
        return OptionalExporter(t)

    def list_extractor(self, t: MetricsExporter[Any]) -> MetricsExporter[Any]:
        return ListExporter(t)

    def dictionary_extractor(
        self,
        key: type,
        value: type,
        key_ext: MetricsExporter[Any],
        val_ext: MetricsExporter[Any]
    ) -> MetricsExporter[Any]:
        if key is str:
            return StringDictionaryExporter(val_ext)
        raise NotImplementedError()

    def enum_extractor(self, enum_name: str, enum_values: List[Tuple[str, Any]]) -> MetricsExporter[Any]:
        return self._enum_exporter
